#include <fstream>
#include <iostream>
//#include "server_socket.h"
#include <sys/socket.h>
#include <vector>
#include <string>
#include <sstream>
#include <cstring>

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <strings.h>

#include <ext/hash_map>
namespace __gnu_cxx {
  template<> struct hash< std::string > {
    size_t operator()( const std::string& x ) const {
      return hash< const char* >()( x.c_str() );
    }
  };
}
typedef __gnu_cxx::hash_map<std::string, int, __gnu_cxx::hash<std::string> > map_string_int;

map_string_int word_lex;
map_string_int pos_lex;
map_string_int gen_lex;
map_string_int num_lex;
map_string_int def_lex;
map_string_int err_lex;

std::vector<std::string> original_text;
std::vector<std::string> inv_err_lex;

int add_or_look_up(map_string_int &lex, std::string str) {
  map_string_int::iterator it = lex.find(str);
  if(it == lex.end()) {
    if(&lex == &err_lex) {
      int res = inv_err_lex.size();
      lex[str] = res;
      inv_err_lex.push_back(str);
      return res;
    } else {
      int res = lex.size();
      lex[str] = res;
      return res;
    }
  } else {
    return it->second;
  }
}

int look_up(map_string_int &lex, std::string str) {
  map_string_int::iterator it = lex.find(str);
  if(it == lex.end()) {
    return -1;
  } else {
    return it->second;
  }
}

typedef int wordtype;
typedef int postype;
typedef int gentype;
typedef int numtype;
typedef int deftype;
typedef int errtype;

struct rule {
  std::vector<std::pair<int, wordtype> > word_constraints;
  std::vector<std::pair<int, postype> > pos_constraints;
  std::vector<std::pair<int, gentype> > gen_constraints;
  std::vector<std::pair<int, numtype> > num_constraints;
  std::vector<std::pair<int, deftype> > def_constraints;
  std::vector<std::pair<int, errtype> > err_constraints;
  errtype newvalue;
};

// fill vector with rules from stream, returns
// maximum offset used in ruleset
int read_rules(std::istream &in, std::vector<rule> &rules) {
  int max_forward_offset = 0;
  int max_backward_offset = 0;
  std::string line;
  std::getline(in, line);
  while(in) {
    if(line[0] != '#') {
      int pos = line.find("RULE:") + 6;
      int end = line.find("=>");
      rule r;
      while(pos < end) {
	int pos2 = line.find("_", pos);
	std::string column = line.substr(pos, pos2-pos);
	
	int offset = atoi(line.c_str() + pos2 + 1);
	if(offset > max_forward_offset)
	  max_forward_offset = offset;
	if(offset < max_backward_offset)
	  max_backward_offset = offset;

	int pos3 = line.find("=", pos2);
	int pos4 = line.find(" ", pos3);
	std::string value = line.substr(pos3+1, pos4-pos3-1);
	if(column == "word") {
	  r.word_constraints.push_back(std::pair<int, wordtype>(offset, add_or_look_up(word_lex,value)));
	  // std::cerr << "word " << offset << " " << add_or_look_up(word_lex,value) << value << "\n";
	}
	else if(column == "pos") {
	  r.pos_constraints.push_back(std::pair<int, postype>(offset, add_or_look_up(pos_lex,value)));
	  // std::cerr << "pos " << offset << " " << add_or_look_up(pos_lex,value) << value << "\n";
	}
	else if(column == "cmp") {
	  r.err_constraints.push_back(std::pair<int, errtype>(offset, add_or_look_up(err_lex,value)));
	  // std::cerr << "cmp " << offset << " " << add_or_look_up(err_lex,value) << value << "\n";
	}
	else if(column == "gen"){
	  r.gen_constraints.push_back(std::pair<int, gentype>(offset, add_or_look_up(gen_lex,value)));
	  // std::cerr << "gen " << offset << " " << add_or_look_up(gen_lex,value) << value << "\n";
	}
	else if(column == "num"){	  
	  r.num_constraints.push_back(std::pair<int, numtype>(offset, add_or_look_up(num_lex,value)));
	  // std::cerr << "num " << offset << " " << add_or_look_up(num_lex,value) << value << "\n";
	}
	else if(column == "def"){
	  r.def_constraints.push_back(std::pair<int, deftype>(offset, add_or_look_up(def_lex,value)));
	  // std::cerr << "def " << offset << " " << add_or_look_up(def_lex,value) << value << "\n";
	}
	pos = pos4 + 1;
      }
      pos = line.find("=", end+3);
      r.newvalue = add_or_look_up(err_lex,line.substr(pos+1));
      //std::cerr << "new val " << line.substr(pos+1) << "\n";
      rules.push_back(r);
    }
    std::getline(in, line);
  }
  if(max_forward_offset > -max_backward_offset)
    return max_forward_offset;
  else
    return - max_backward_offset;
}

void slow_apply(rule &r, 
		std::vector<wordtype> &words,
		std::vector<postype> &poss,
		std::vector<errtype> &errs,
		std::vector<gentype> &gens,
		std::vector<numtype> &nums,
		std::vector<deftype> &defs
		) {
  int ZZZ = look_up(word_lex,"ZZZ");
  for(int i = 0; i < words.size(); i++) {
    if(words[i] != ZZZ) {
      int applies = 1;
      for(int j = 0; j < r.word_constraints.size(); j++)
	if(i + r.word_constraints[j].first < 0 
	   || i + r.word_constraints[j].first >= words.size()
	   || words[i + r.word_constraints[j].first] != r.word_constraints[j].second) {
	  applies = 0;
	  break;
	}
      for(int j = 0; applies && j < r.pos_constraints.size(); j++)
	if(i + r.pos_constraints[j].first < 0 
	   || i + r.pos_constraints[j].first >= poss.size()
	   || poss[i + r.pos_constraints[j].first] != r.pos_constraints[j].second) {
	  applies = 0;
	  break;
	}
      for(int j = 0; applies && j < r.gen_constraints.size(); j++)
	if(i + r.gen_constraints[j].first < 0 
	   || i + r.gen_constraints[j].first >= gens.size()
	   || gens[i + r.gen_constraints[j].first] != r.gen_constraints[j].second) {
	  applies = 0;
	  break;
	}
      for(int j = 0; applies && j < r.num_constraints.size(); j++)
	if(i + r.num_constraints[j].first < 0 
	   || i + r.num_constraints[j].first >= nums.size()
	   || nums[i + r.num_constraints[j].first] != r.num_constraints[j].second) {
	  applies = 0;
	  break;
	}
      for(int j = 0; applies && j < r.def_constraints.size(); j++)
	if(i + r.def_constraints[j].first < 0 
	   || i + r.def_constraints[j].first >= defs.size()
	   || defs[i + r.def_constraints[j].first] != r.def_constraints[j].second) {
	  applies = 0;
	  break;
	}
      for(int j = 0; applies && j < r.err_constraints.size(); j++)
	if(i + r.err_constraints[j].first < 0 
	   || i + r.err_constraints[j].first >= errs.size()
	   || errs[i + r.err_constraints[j].first] != r.err_constraints[j].second) {
	  applies = 0;
	  break;
	}
      if(applies) {
	errs[i] = r.newvalue;	
      }
    }
  }
}

/* Assuming the text is Swedish text, this should return 1 if the
   encoding is UTF-8, and it should return 0 if we find something that
   looks like latin-1 */
int looksLikeUTF(const char *s) {
  if(s == NULL) {
    return 1;
  }

  int last = 1;
  for(int i = 0; s[i] != '\0'; i++) {
    if(s[i] < 0 && last > 0) {
      if(s[i] == -27
	 || s[i] == -28
	 || s[i] == -10
	 || s[i] == -59
	 || s[i] == -60
	 || s[i] == -42) {
	return 0;
      }
    }
    if(s[i] < 0 && last == -61) {
      return 1;
    }
    last = s[i];
  }
  
  return 1;
}

/* Converts UTF-8 to Latin-1 correctly if we assume the string
   contents are Swedish text. Based on code example from Mark Ransom */
void utf2latin1(std::string &s) {
  
  int n = s.size();
  std::string temp = "";
  temp.reserve(n);

  unsigned int codepoint = 0;

  int j = 0;
  for (int i = 0; s[i] != 0; ) {
    /* unsigned char c = static_cast<unsigned char>(s[i]); */
    unsigned char c = (unsigned char)(s[i]);
    if (c <= 0x7f)
      codepoint = c;
    else if (c <= 0xbf)
      codepoint = (codepoint << 6) | (c & 0x3f);
    else if (c <= 0xdf)
      codepoint = c & 0x1f;
    else if (c <= 0xef)
      codepoint = c & 0x0f;
    else
      codepoint = c & 0x07;

    i++;
    
    if (((s[i] & 0xc0) != 0x80) && (codepoint <= 0x10ffff)) {
      if (codepoint <= 255) {
	/* tempBuf[j++] = static_cast<char>(codepoint); */
	temp += (char)(codepoint);
      } else {
	// do whatever you want for out-of-bounds characters
	temp += 'X'; // Store something away so we can convert back later? TODO: Jonas
      }
    }
  }

  s = temp;
}

const int SOCKETBUFSIZE = 1000000;
void read_data(int socket,
	       std::vector<wordtype> &words,
	       std::vector<postype> &poss,
	       std::vector<errtype> &errs,
	       std::vector<gentype> &gens,
	       std::vector<numtype> &nums,
	       std::vector<deftype> &defs,
	       int max_offset
	       ) {
  char buf[SOCKETBUFSIZE+1];
  buf[0] = 0;
  bool done = 0;
  std::string text = "";

  while(!done) {
    int len = 0, offs = 0;
    while(offs < SOCKETBUFSIZE && !done) {
      len = read(socket, buf+offs, SOCKETBUFSIZE - offs);
      if(len > 0) {
	buf[SOCKETBUFSIZE] = 0;
	char *end = strstr(buf, "ENDQ");
	if(end) {
	  *end = 0;
	  done = 1;
	} else {
	  offs += len;
	}
      } else {
	buf[offs] = 0;
	done = 1;
      }

      if(offs >= SOCKETBUFSIZE) {
	buf[SOCKETBUFSIZE-1] = 0;
	done = 1;
      }
    }
    text += buf;
    int tl = text.length() - 10;
    if(tl < 0)
      tl = 0;
    if(text.substr(tl).find("ENDQ") != std::string::npos) {
      done = 1;
    }
  }

  if(looksLikeUTF(text.c_str())) {
    utf2latin1(text);
  }

  std::istringstream in(text);

  std::string line;

  int MAD = add_or_look_up(pos_lex, "mad");

  for(int i = 0; i < max_offset; i++) {
    words.push_back(add_or_look_up(word_lex,"ZZZ"));
    original_text.push_back("ZZZ");
    poss.push_back(add_or_look_up(pos_lex,"ZZZ"));
    gens.push_back(add_or_look_up(gen_lex,"ZZZ"));
    nums.push_back(add_or_look_up(num_lex,"ZZZ"));
    defs.push_back(add_or_look_up(def_lex,"ZZZ"));
    errs.push_back(add_or_look_up(err_lex,"ZZZ"));
  }

  while(in && line.substr(0,4) != "ENDQ") {
    std::getline(in, line);
    if(line != "" && line.substr(0,4) != "ENDQ") {
      int pos = line.find_first_of(" \t");
      words.push_back(look_up(word_lex,line.substr(0,pos)));
      original_text.push_back(line.substr(0,pos));

      int pos2 = line.find_first_of(" \t\n\r", pos+1);
      if(pos2 == std::string::npos) {
	poss.push_back(add_or_look_up(pos_lex,line.substr(pos+1)));
	pos2 = line.length();
      } else {
	poss.push_back(add_or_look_up(pos_lex,line.substr(pos+1,pos2-pos-1)));
      }

      pos = pos2;
      pos2 = line.find_first_of(" \t\n\r", pos+1);
      if(pos2 == std::string::npos) {
	gens.push_back(add_or_look_up(gen_lex,line.substr(pos+1)));
	pos2 = line.length();
      } else {
	gens.push_back(add_or_look_up(gen_lex,line.substr(pos+1,pos2-pos-1)));
      }
      pos = pos2;
      pos2 = line.find_first_of(" \t\n\r", pos+1);
      if(pos2 == std::string::npos) {
	nums.push_back(add_or_look_up(num_lex,line.substr(pos+1)));
	pos2 = line.length();
      } else {
	nums.push_back(add_or_look_up(num_lex,line.substr(pos+1,pos2-pos-1)));
      }
      pos = pos2;
      pos2 = line.find_first_of(" \t\n\r", pos+1);
      if(pos2 == std::string::npos) {
	defs.push_back(add_or_look_up(def_lex,line.substr(pos+1)));
	pos2 = line.length();
      } else {
	defs.push_back(add_or_look_up(def_lex,line.substr(pos+1,pos2-pos-1)));
      }

      int pos3 = line.find_first_of(" \t\r\n", pos2+1);
      if(pos3 != std::string::npos) {
	// if there is no column with this data, pos2 is 
	// normaly line.length() so we get errs[i] = "", 
	// which is fine
	errs.push_back(add_or_look_up(err_lex,line.substr(pos2+1, pos3-pos2-1)));
      } else if(pos2 + 1 < line.length()) {
	errs.push_back(add_or_look_up(err_lex,line.substr(pos2+1)));
      } else {
	errs.push_back(add_or_look_up(err_lex,""));
      }
      
      if(poss[poss.size()-1] == MAD) {
	for(int i = 0; i < max_offset; i++) {
	  words.push_back(add_or_look_up(word_lex,"ZZZ"));
	  original_text.push_back("ZZZ");
	  poss.push_back(add_or_look_up(pos_lex,"ZZZ"));
	  gens.push_back(add_or_look_up(gen_lex,"ZZZ"));
	  nums.push_back(add_or_look_up(num_lex,"ZZZ"));
	  defs.push_back(add_or_look_up(def_lex,"ZZZ"));
	  errs.push_back(add_or_look_up(err_lex,"ZZZ"));
	}
      }
    }
  }
  
  for(int i = 0; i < max_offset; i++) {
    words.push_back(add_or_look_up(word_lex,"ZZZ"));
    original_text.push_back("ZZZ");
    poss.push_back(add_or_look_up(pos_lex,"ZZZ"));
    gens.push_back(add_or_look_up(gen_lex,"ZZZ"));
    nums.push_back(add_or_look_up(num_lex,"ZZZ"));
    defs.push_back(add_or_look_up(def_lex,"ZZZ"));
    errs.push_back(add_or_look_up(err_lex,"ZZZ"));
  }
}

void initial_guess(
		   std::vector<wordtype> &words,
		   std::vector<postype> &poss,
		   std::vector<errtype> &errs,
		   std::vector<gentype> &gens,
		   std::vector<numtype> &nums,
		   std::vector<deftype> &defs
		   ) {
  int ZZZ = look_up(word_lex,"ZZZ");
  int OK = add_or_look_up(err_lex,"OK");

  for(int i = 0; i < words.size(); i++) {
    if(words[i] != ZZZ) {
      errs[i] = OK;
    }
  }
}

// taken from a socket primer
int establish(int portnum)
{ 
  const int MAXHOSTNAME = 1000;
  char myname[MAXHOSTNAME+1];
  int s;
  struct sockaddr_in sa;
  struct hostent *hp;
  
  bzero(&sa,sizeof(struct sockaddr_in));
  gethostname(myname,MAXHOSTNAME);
  hp= gethostbyname(myname);
  if (hp == NULL)
    return -1;
  sa.sin_family= hp->h_addrtype; /* this is our host address */
  sa.sin_port= htons(portnum); /* this is our port number */
  if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
    return -1;
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa)) < 0) {
    close(s);
    return -1;
  }
  listen(s, 5); /* max # of queued connects */
  return s;
}

// taken from a socket primer
int get_connection(int s // socket
		   )
{ 
  struct sockaddr_in isa;
  socklen_t i;
  int t;
  
  i = sizeof(isa);
  getsockname(s,(struct sockaddr *)&isa,&i);
  
  if ((t = accept(s,(struct sockaddr *)&isa,&i)) < 0)   /* accept connection if there is one */
    return -1;
  return t;
}

int main(int argc, char **argv) {

  if(argc < 3) {
    std::cerr << argv[0] << " rule-file port" << std::endl;
    return 0;
  }
  /* handler for child processes */
  signal(SIGCLD, SIG_IGN);
  int portno = atoi(argv[2]);
  int server_socket;
  if ((server_socket = establish(portno)) < 0) {  /* plug in the phone */
    std::cerr << "could not open socket for listening\n";
    exit(1);
  }

  std::vector<rule> rules;
  std::ifstream rulefile(argv[1]);
  int max_offset = read_rules(rulefile, rules);
  if(max_offset < 1)
    max_offset = 1;

  std::vector<wordtype> words;
  std::vector<postype> poss;
  std::vector<errtype> errs;
  std::vector<gentype> gens;
  std::vector<numtype> nums;
  std::vector<deftype> defs;

  while(1) {
    std::cerr << "waiting...\n";
    words.clear();
    poss.clear();
    errs.clear();
    gens.clear();
    nums.clear();
    defs.clear();
    original_text.clear();

    int client_socket = 0;

    if ((client_socket = get_connection(server_socket)) < 0) {
      if (errno == EINTR)
	continue;
      std::cerr << "error on accept\n";
      exit(1);
    }
    
    // fork all client calls, that way, if they hang or crash,
    // the server keeps working
    std::ostringstream buf;
    switch(fork()) {
    case -1 : // error
      std::cerr << "error on fork()\n";
      exit(1);
    case 0 : // child
      close(server_socket);
      alarm(60*2); // if we still haven't finished in 2 minutes, die...
      read_data(client_socket,
		words,
		poss,
		errs,
		gens,
		nums,
		defs,
		max_offset
		);
      alarm(0); // remove alarm
      initial_guess(
		    words,
		    poss,
		    errs,
		    gens,
		    nums,
		    defs
		    );
      for(int i = 0; i < rules.size(); i++)
	slow_apply(rules[i], 
		   words,
		   poss,
		   errs,
		   gens,
		   nums,
		   defs
		   );
      buf << "<reply>\n";
      for(int i = 0; i < words.size(); i++)
	if(words[i] != look_up(word_lex,"ZZZ"))
	  buf << original_text[i] << "\t" << inv_err_lex[errs[i]] << "\n";
      buf << "</reply>\n";
      write(client_socket, buf.str().c_str(), buf.str().length() * sizeof(char));
      
      close(client_socket);
      exit(0); // child is finished      
    default : // parent
      close(client_socket);
      wait(NULL); // to avoid spamming lots of processes, wait for child to finish
    }    
  }
  return 0;
}
